from abc import ABC, abstractmethod


class TaskEntryPoint(ABC):
    """
    An abstract base class that acts as an entry point for tasks.

    This class is meant to be subclassed and extended with specific implementation
    details. The subclass must implement
    a method `task()` which serves as the entry point for the task logic.

    Usage:
        To use `TaskEntryPoint`, create a subclass that implements the `task()` method,
        and call it as follows:

        class MyTask(TaskEntryPoint):
            @classmethod
            def task(cls) -> None:
                # implement the task logic here

        MyTask.task()

    Raises:
        NotImplementedError: If `task()` method is not implemented in the subclass.
    """

    @classmethod
    @abstractmethod
    def task(cls) -> None:
        """
        Abstract class method that serves as the entry point for the task logic.
        This method should be implemented by the subclass with the specific
        implementation details of the task.
        """
        raise NotImplementedError()



import re
import unittest
from datetime import datetime, timedelta

import pyspark.sql.functions as F
from pyspark.sql import DataFrame
from pyspark.sql.types import IntegerType, StructField, StructType
from spetlrtools.testing import DataframeTestCase

from spetlr.etl.log import LogTransformer
from spetlr.etl.types import dataset_group
from spetlr.utils import DataframeCreator


class LogTransformerSubclass(LogTransformer):
    def log(self, df: DataFrame) -> DataFrame:
        return df.select(F.count("*").alias("Count"))

    def log_many(self, dataset: dataset_group) -> DataFrame:
        return super().log_many(dataset)


class TestLogTransformer(DataframeTestCase):
    @classmethod
    def setUpClass(cls) -> None:
        cls.df_empty = DataframeCreator.make(schema=StructType(), data=[()])

        schema = StructType(
            [
                StructField("col_1", IntegerType(), True),
                StructField("col_2", IntegerType(), True),
            ]
        )

        data = [
            (1, 1),
            (2, None),
            (3, 3),
            (4, None),
        ]

        df_with_data_1 = DataframeCreator.make(schema, data)
        df_with_data_2 = DataframeCreator.make(schema, data)

        cls.dataset = {
            "df_1": df_with_data_1,
            "df_2": df_with_data_2,
        }

    def test_log_is_abstractmethod_01(self) -> None:
        self.assertTrue(getattr(LogTransformer.log, "__isabstractmethod__"))

    def test_log_many_is_abstractmethod_02(self) -> None:
        self.assertTrue(getattr(LogTransformer.log_many, "__isabstractmethod__"))

    def test_log_many_default_behaviour_03(self) -> None:
        df = LogTransformerSubclass(log_name="test").log_many(self.dataset)

        self.assertDataframeMatches(
            df=df,
            columns=["Count", "DatasetInputKey"],
            expected_data=[
                (4, "df_1"),
                (4, "df_2"),
            ],
        )

    def test_create_log_04(self) -> None:
        ltsb = LogTransformerSubclass(
            log_name="test_log_name",
        )

        df = ltsb._create_log(self.dataset["df_1"])

        self.assertDataframeMatches(
            df=df,
            columns=[
                # "LogId", tested separately because uuid
                "LogName",
                # "LogTimestamp", tested separately because datetime
                "LogMethodName",
                "col_1",
                "col_2",
            ],
            expected_data=[
                ("test_log_name", "LogTransformerSubclass", 1, 1),
                ("test_log_name", "LogTransformerSubclass", 2, None),
                ("test_log_name", "LogTransformerSubclass", 3, 3),
                ("test_log_name", "LogTransformerSubclass", 4, None),
            ],
        )

        row_1 = df.collect()[0]

        log_id = row_1[0]
        pattern = (
            r"^[0-9a-f]{8}-[0-9a-f]{4}-4[0-9a-f]{3}-[89ab][0-9a-f]{3}-[0-9a-f]{12}$"
        )
        self.assertTrue(re.match(pattern, log_id))

        log_timestamp = row_1[2]
        self.assertTrue(isinstance(log_timestamp, datetime))

        datetime_now = datetime.utcnow()
        datetime_now_minus_10_minutes = datetime_now - timedelta(minutes=10)
        datetime_now_plus_10_minutes = datetime_now + timedelta(minutes=10)
        self.assertTrue(
            datetime_now_minus_10_minutes < log_timestamp < datetime_now_plus_10_minutes
        )

    def test_process_05(self) -> None:
        df = LogTransformerSubclass(
            log_name="test_log_name",
            dataset_input_keys=["df_key"],
        ).process(self.dataset["df_1"])

        dataset_input_key_generated = df.select("DatasetInputKey").collect()[0][0]
        self.assertEqual(dataset_input_key_generated, "df_key")

    def test_process_many_06(self) -> None:
        pass
        # has not logic to test as 'log_many' and '_create_log' has already been tested


if __name__ == "__main__":
    unittest.main()



from typing import List

from spetlr.delta import DeltaHandle
from spetlr.etl import EtlBase, Orchestrator
from spetlr.etl.extractors import IncrementalExtractor, SimpleExtractor
from spetlr.etl.loaders import SimpleLoader
from spetlr.exceptions import MissingUpsertJoinColumns
from spetlr.orchestrators.ehjson2delta.EhJsonToDeltaTransformer import (
    EhJsonToDeltaTransformer,
)


class EhToDeltaSilverOrchestrator(Orchestrator):
    """
    This class has been designed to carry out the ETL task
    of unpacking and transforming bronze eventhub data to the silver layer.

    Parameters:

    dh_source: DeltaHandle for the source delta table (bronze)
    dh_target: DeltaHandle for the target delta table (silver)
    upsert_join_cols: The columns used for upserting.
    mode: The mode of data load (upsert, append, overwrite).

    Returns:
    Processed datasets of the super Orchestrator class
    """

    def __init__(
        self,
        dh_source: DeltaHandle,
        dh_target: DeltaHandle,
        upsert_join_cols: List[str] = None,
        mode: str = "upsert",
    ):
        super().__init__()
        self.dh_source = dh_source
        self.dh_target = dh_target
        self.mode = mode
        self.upsert_join_cols = upsert_join_cols

        # step 1
        # Extracts the data from the bronze layer
        if mode == "upsert" or mode == "append":
            if mode == "upsert":
                if upsert_join_cols is None:
                    raise MissingUpsertJoinColumns
            self.extract_from(
                IncrementalExtractor(
                    handle_source=self.dh_source,
                    handle_target=self.dh_target,
                    time_col_source="EnqueuedTimestamp",
                    time_col_target="EnqueuedTimestamp",
                )
            )

        else:
            self.extract_from(SimpleExtractor(dh_source, "EhDeltaBronze"))

        # step 2,
        #  - use the target schema to select what to copy from capture files
        #  - anything that is not in the source df is used to unpack the body json
        self.transform_with(EhJsonToDeltaTransformer(target_dh=dh_target))

        # the method filter_with can be used to insert any number of transformers here

        self._loader = SimpleLoader(
            dh_target, join_cols=self.upsert_join_cols, mode=self.mode
        )

        self.load_into(self._loader)

    @classmethod
    def from_tc(cls, dh_source_id: str, dh_target_id: str):
        return cls(
            dh_source=DeltaHandle.from_tc(dh_source_id),
            dh_target=DeltaHandle.from_tc(dh_target_id),
        )

    def filter_with(self, etl: EtlBase):
        """Additional filters to execute before loading."""
        loader = self.steps.pop()

        # the following will fail if additional steps have been added.
        assert self._loader is loader, "unexpected change in etl steps"

        self.transform_with(etl)
        self.load_into(loader)



import pyspark.sql.types as T
from spetlrtools.testing import DataframeTestCase

from spetlr.spark import Spark
from spetlr.transformers import UnionTransformer


class TestUnionTransformer(DataframeTestCase):
    def test_union_two_dataframes(self):
        input_schema = T.StructType(
            [
                T.StructField("Col1", T.StringType(), True),
                T.StructField("Col2", T.IntegerType(), True),
                T.StructField("Col3", T.DoubleType(), True),
                T.StructField("Col4", T.StringType(), True),
                T.StructField("Col5", T.StringType(), True),
            ]
        )

        input_data1 = [("Col1Data", 42, 13.37, "Col4Data", "Col5Data")]
        input_data2 = [("Col1Data_2nd", 42, 13.37, "Col4Data_2nd", "Col5Data_2nd")]

        input_df_1 = Spark.get().createDataFrame(data=input_data1, schema=input_schema)
        input_df_2 = Spark.get().createDataFrame(data=input_data2, schema=input_schema)
        input_datasets = {"DF1": input_df_1, "DF2": input_df_2}

        transformed_df = UnionTransformer().process_many(input_datasets)

        expected_data = [
            ("Col1Data", 42, 13.37, "Col4Data", "Col5Data"),
            ("Col1Data_2nd", 42, 13.37, "Col4Data_2nd", "Col5Data_2nd"),
        ]

        self.assertDataframeMatches(
            df=transformed_df,
            expected_data=expected_data,
        )

    def test_union_three_dataframes(self):
        input_schema = T.StructType(
            [
                T.StructField("Col1", T.StringType(), True),
                T.StructField("Col2", T.IntegerType(), True),
                T.StructField("Col3", T.DoubleType(), True),
                T.StructField("Col4", T.StringType(), True),
                T.StructField("Col5", T.StringType(), True),
            ]
        )

        input_data1 = [("Col1Data", 42, 13.37, "Col4Data", "Col5Data")]
        input_data2 = [("Col1Data_2nd", 42, 13.37, "Col4Data_2nd", "Col5Data_2nd")]
        input_data3 = [
            (
                "Totally new data",
                1234,
                98.76,
                "More totally new data",
                "Again, totally new data",
            )
        ]

        input_df_1 = Spark.get().createDataFrame(data=input_data1, schema=input_schema)
        input_df_2 = Spark.get().createDataFrame(data=input_data2, schema=input_schema)
        input_df_3 = Spark.get().createDataFrame(data=input_data3, schema=input_schema)
        input_datasets = {"DF1": input_df_1, "DF2": input_df_2, "DF3": input_df_3}

        transformed_df = UnionTransformer().process_many(input_datasets)

        expected_data = [
            ("Col1Data", 42, 13.37, "Col4Data", "Col5Data"),
            ("Col1Data_2nd", 42, 13.37, "Col4Data_2nd", "Col5Data_2nd"),
            (
                "Totally new data",
                1234,
                98.76,
                "More totally new data",
                "Again, totally new data",
            ),
        ]

        self.assertDataframeMatches(df=transformed_df, expected_data=expected_data)

    def test_union_dataframes_missing_columns(self):
        input_schema1 = T.StructType(
            [
                T.StructField("Col1", T.StringType(), True),
                T.StructField("Col2", T.IntegerType(), True),
                T.StructField("Col3", T.DoubleType(), True),
                T.StructField("Col4", T.StringType(), True),
                T.StructField("Col5", T.StringType(), True),
            ]
        )
        input_data1 = [
            (
                "Col1Data",
                42,
                13.37,
                "Col4Data",
                "Col5Data",
            )
        ]

        input_schema2 = T.StructType(
            [
                T.StructField("Col1", T.StringType(), True),
                T.StructField("Col2", T.IntegerType(), True),
                T.StructField("Col3", T.DoubleType(), True),
                T.StructField("Col4", T.StringType(), True),
            ]
        )
        input_data2 = [("Col1Data_2nd", 42, 13.37, "Col4Data_2nd")]

        input_df_1 = Spark.get().createDataFrame(data=input_data1, schema=input_schema1)
        input_df_2 = Spark.get().createDataFrame(data=input_data2, schema=input_schema2)
        input_datasets = {"DF1": input_df_1, "DF2": input_df_2}

        transformed_df = UnionTransformer(allowMissingColumns=True).process_many(
            input_datasets
        )

        expected_data = [
            ("Col1Data", 42, 13.37, "Col4Data", "Col5Data"),
            ("Col1Data_2nd", 42, 13.37, "Col4Data_2nd", None),
        ]

        self.assertDataframeMatches(df=transformed_df, expected_data=expected_data)



from datetime import datetime as dt
from datetime import timezone
from typing import List, Optional

from pyspark.sql import DataFrame
from pyspark.sql import functions as f

from spetlr.delta import DeltaHandle
from spetlr.eh.EventHubCaptureExtractor import EventHubCaptureExtractor
from spetlr.etl import Extractor
from spetlr.exceptions import EhJsonToDeltaException
from spetlr.spark import Spark


class EhJsonToDeltaExtractor(Extractor):
    """
    This extractor has a side effect on the delta table!

    Get the highest previously read partition from the delta table.
    Truncate all rows from the delta table that come from that partition.
    Extract all rows from the EventHubCaptureExtractor from that partition on.
    """

    def __init__(
        self, eh: EventHubCaptureExtractor, dh: DeltaHandle, dataset_key: str = None
    ):
        super().__init__(dataset_key=dataset_key)
        self.eh = eh
        self.dh = dh

    def _read_pdate_partitioned(self) -> DataFrame:
        """get the highest pdate partition,
        truncate it,
        read the event hub from that partition"""
        pdate_df = self.dh.read().select("pdate")

        # assert correct schema
        assert pdate_df.schema.fields[0].dataType.typeName().upper() == "TIMESTAMP"

        max_pdate: Optional[dt] = pdate_df.groupBy().agg(f.max("pdate")).collect()[0][0]
        if max_pdate is None:
            # if it is None, no previous data exists,
            # so don't truncate and read everything
            return self.eh.read()

        max_pdate = max_pdate.astimezone(timezone.utc)

        # truncate this largest partition...
        self._delete_table_partitions(
            self.dh.get_tablename(), [f"pdate='{max_pdate.isoformat()}'"]
        )

        # the datetime literal specification:
        # https://spark.apache.org/docs/latest/sql-ref-literals.html#datetime-literal
        # ...it works with the python datetime .isoformat()

        # ...and read it back from eventhub
        return self.eh.read(from_partition=max_pdate)

    def _delete_table_partitions(self, tbl_name: str, conditions: List[str]) -> None:
        # My first attempt was TRUNCATE TABLE table PARTITION (<spec>),
        # but it turns out that this is newer than spark 3.1 which we are using.
        # Second attempt was to ALTER TABLE table DROP PARTITION (<spec>),
        # but I got the error that `ALTER TABLE DROP PARTITION` is not supported
        # for Delta tables

        # So finally this uses a DELETE FROM in the hope that Spark can optimize this
        # to efficiently remove the partition
        Spark.get().sql(f"DELETE FROM {tbl_name} WHERE {' AND '.join(conditions) }")

    def _read_ymd_ymdh_partitioned(self) -> DataFrame:
        """get the highest partition, piece by piece,
        truncate it,
        construct the datetime that the pieces correspond to,
        read the event hub from that datetime"""

        dh_parts = self.dh.get_partitioning()

        # this df will be filtered stepwise
        df = self.dh.read()

        # assert partitioning columns are all integers
        for field in df.select(*dh_parts).schema.fields:
            assert field.dataType.typeName().upper() == "INTEGER"

        y = df.groupBy().agg(f.max("y")).collect()[0][0]
        if y is None:
            # if it is None, no previous data exists,
            # so don't truncate and read everything
            return self.eh.read()

        # continue the logic. We know there is a partition.
        truncate_partiton_spec = [f"y={y}"]
        df = df.filter(f"y={y}")

        # if there was a y, there is a partition and hence the others must exist
        m = df.groupBy().agg(f.max("m")).collect()[0][0]
        truncate_partiton_spec.append(f"m={m}")
        df = df.filter(f"m={m}")

        d = df.groupBy().agg(f.max("d")).collect()[0][0]
        truncate_partiton_spec.append(f"d={d}")
        df = df.filter(f"d={d}")

        if "h" in dh_parts:
            h = df.groupBy().agg(f.max("h")).collect()[0][0]
            truncate_partiton_spec.append(f"h={h}")
        else:
            h = 0

        self._delete_table_partitions(self.dh.get_tablename(), truncate_partiton_spec)

        read_from = dt(y, m, d, h, tzinfo=timezone.utc)

        return self.eh.read(from_partition=read_from)

    def read(self) -> DataFrame:
        # we need to find out 2 things,
        # - what (if any) to truncate from the delta table,
        # - and where to read the eventhub from (or ead everything)

        # first check if the partitioning is usable
        eh_parts = self.eh.get_partitioning()
        dh_parts = self.dh.get_partitioning()

        if dh_parts == ["pdate"]:
            return self._read_pdate_partitioned()
        if dh_parts == eh_parts:
            # its ymd or ymdh
            return self._read_ymd_ymdh_partitioned()
        else:
            raise EhJsonToDeltaException("Delta table has bad partitioning")



import unittest
from datetime import datetime

from pyspark.sql.types import (
    BooleanType,
    IntegerType,
    StringType,
    StructField,
    StructType,
    TimestampType,
)

from spetlr.azure_log_analytics import AzureLogAnalyticsHandle
from spetlr.functions import init_dbutils
from spetlr.spark import Spark


class TestAzureLogAnalyticsHandle(unittest.TestCase):
    def test_api_post(self):
        # fetching secrets that have been created by the deployment pipeline
        workspace_id = init_dbutils().secrets.get(
            "secrets", "LogAnalyticsWs--WorkspaceID"
        )

        primary_key = init_dbutils().secrets.get(
            "secrets", "LogAnalyticsWs--PrimaryKey"
        )

        az_lah = AzureLogAnalyticsHandle(
            log_analytics_workspace_id=workspace_id,
            shared_key=primary_key,
        )

        schema = StructType(
            [
                StructField("col_1", StringType(), True),
                StructField("col_2", IntegerType(), True),
                StructField("col_3", BooleanType(), True),
                StructField("col_4", TimestampType(), True),
            ]
        )

        data = [
            (
                "value",
                2903,
                True,
                datetime(2020, 2, 2),
            )
        ]

        spark = Spark().get()

        df = spark.createDataFrame(data, schema)

        status_code = az_lah.api_post(df)

        self.assertEqual(
            first=status_code,
            second=200,
        )



from typing import Any, List, Union

from pyspark.sql import DataFrame

from spetlr.etl import Loader, dataset_group
from spetlr.tables import TableHandle


class DeleteDataLoader(Loader):
    def __init__(
        self,
        handle: TableHandle,
        comparison_col: str,
        comparison_limit: Any,
        comparison_operator: str = "<",
        dataset_input_keys: Union[str, List[str]] = None,
    ):
        super().__init__(dataset_input_keys=dataset_input_keys)

        self.handle = handle
        self.comparison_col = comparison_col
        self.comparison_limit = comparison_limit
        self.comparison_operator = comparison_operator

    def save_many(self, datasets: dataset_group) -> None:
        self.save(None)

    def save(self, df: DataFrame) -> None:
        """Delete old data from a given table."""
        self.handle.delete_data(
            self.comparison_col, self.comparison_limit, self.comparison_operator
        )



import re
from typing import Dict

from spetlr.configurator.sql.exceptions import _UnpackAttemptFailed
from spetlr.configurator.sql.StatementBlocks import StatementBlocks
from spetlr.configurator.sql.substructures import _unpack_options
from spetlr.configurator.sql.types import _PeekableTokenList
from spetlr.exceptions.configurator_exceptions import (
    SpetlrConfiguratorInvalidSqlException,
)


def _walk_db_statement(statement: _PeekableTokenList) -> Dict:
    # for databases, we need to collect all tokens until the next database block matches
    name_parts = []
    blocks = StatementBlocks()
    while True:
        # until we managed to unpack the rest of the database,
        # all the bits must be part of the name.
        # This allows unpacking names like CREATE SCHEMA {baseName}mydb{ID}
        # which will return lots of parsing errors otherwise.

        name_parts.append(next(statement).value.strip("`"))

        try:
            blocks = _extract_db_blocks(statement)
        except _UnpackAttemptFailed:
            continue
        break

    blocks.using = "db"

    # all validation is completed.
    # construct the configurator object

    object_details = blocks.get_simple_structure()
    object_details["name"] = "".join(name_parts)

    return object_details


def _extract_db_blocks(stmt: _PeekableTokenList) -> StatementBlocks:
    """After the initial "create table [if not exists] mytable",
    The statements can contain several blocks that are mostly optional,
    (check the documentation links at the top of this file)
    and which can come in any order.
    (we are not validating SQL here, only extracting information)
    This function extacts those blocks and makes them available.
    """
    blocks = StatementBlocks()
    try:
        while True:
            try:
                val = stmt.peek().value.upper()
            except StopIteration:
                break

            if val == "LOCATION":
                next(stmt)
                blocks.location = next(stmt).value.strip("\"'")
                continue

            if val == "COMMENT":
                next(stmt)
                blocks.comment = next(stmt).value.strip("\"'")
                continue

            if re.match(r"WITH\s+DBPROPERTIES", val):
                next(stmt)
                blocks.dbproperties = {}
                for k, v in _unpack_options(stmt):
                    blocks.dbproperties[k] = v
                continue

            if val == ";":
                next(stmt)
                break

            # if we got to here with no exception, break or continue something is wrong
            raise _UnpackAttemptFailed("Unknown statement form encountered.")
    except _UnpackAttemptFailed as e:
        if blocks == StatementBlocks():
            # only raise this exception if no token has been consumed.
            raise
        else:
            raise SpetlrConfiguratorInvalidSqlException(*(e.args))
    return blocks



import unittest

from spetlr.entry_points.task_entry_point import TaskEntryPoint


class TestModuleHelper(unittest.TestCase):
    def test_task_as_abstract_method(self):
        self.assertTrue(hasattr(TaskEntryPoint.task, "__isabstractmethod__"))


if __name__ == "__main__":
    unittest.main()



from typing import List

import pyspark.sql.functions as f
from pyspark.sql import DataFrame
from pyspark.sql.window import Window


def DropOldestDuplicates(
    *, df: DataFrame, cols: List[str], orderByColumn: str
) -> DataFrame:
    """
    Removes the oldest duplicates.

    In other words, if there is multiple duplicates, only the newest row remain.

    param df: The pyspark DataFrame with potential duplicates
    param cols: A list of column names that the window function is partitioned by
    param orderByColumn: The time formatted column to order by

    return: DataFrame with oldest duplicate rows removed

    """
    wnd = Window.partitionBy(cols).orderBy(df[orderByColumn].desc())
    df = df.withColumn("rowNum", f.row_number().over(wnd)).where(f.col("rowNum") == 1)
    df = df.drop("rowNum")
    return df



import json
from string import Formatter
from typing import Any, Dict

import pyspark.sql.types as T

from spetlr.configurator import Configurator
from spetlr.exceptions import (
    FalseSchemaDefinitionException,
    NoSuchSchemaException,
    NoSuchValueException,
    UnregisteredSchemaDefinitionException,
)
from spetlr.schema_manager.spark_schema import get_schema
from spetlr.singleton import Singleton


class SchemaManager(metaclass=Singleton):
    _DEFAULT = object()
    # This dict contains the registered schemas
    _registered_schemas: Dict[str, T.StructType] = {}

    def __init__(self):
        self.clear_all_configurations()

    def clear_all_configurations(self):
        self._registered_schemas = dict()

    def register_schema(self, schema_name: str, schema: T.StructType) -> None:
        self._registered_schemas[schema_name] = schema

    def get_schema(
        self,
        schema_identifier: str,
        default: Any = _DEFAULT,
    ) -> T.StructType:
        """
        Get a schema from either the registered schemas or the tables available
        to the Configurator.\n
        \"schema_identifier\" accepts either the name of a registered schema or
        a Configurator table id.
        """

        # Check if the identifier was directly registered or previously retrieved
        if schema_identifier in self._registered_schemas.keys():
            return self._registered_schemas[schema_identifier]

        # Otherwise, the schema identifier must be table identifier
        try:
            schema = Configurator().get(table_id=schema_identifier, property="schema")
        except NoSuchValueException:
            if default is self._DEFAULT:
                raise NoSuchSchemaException(schema_identifier)
            else:
                return default

        # If the schema is a string, look it up as another schema
        if isinstance(schema, str):
            # exceptions raised here will correctly roll up to the caller.
            # recursive call allows for stacked definitions
            other_schema = self.get_schema(schema)
            self._registered_schemas[schema_identifier] = other_schema
            return other_schema

        # Otherwise, it must be a dict
        if not isinstance(schema, dict):
            raise UnregisteredSchemaDefinitionException(schema)

        # Ensure that the dict has exactly 1 item
        [(key, value)] = schema.items()

        # Check if the schema is a sql string
        if key == "sql":
            # Check for substitutions like {} contained in the value:
            other_schema_keys = set(
                i[1] for i in Formatter().parse(value) if i[1] is not None
            )
            replacements = {}
            for other_key in other_schema_keys:
                if other_key.endswith("_schema"):
                    other_base_key = other_key[: -len("_schema")]
                    replacements[other_key] = self.get_schema_as_string(other_base_key)
                else:
                    replacements[other_key] = Configurator().get_all_details()[
                        other_key
                    ]
            if replacements:
                value = value.format(**replacements)

            parsed_schema = get_schema(value)
            if not isinstance(parsed_schema, T.StructType):
                raise FalseSchemaDefinitionException()
            self._registered_schemas[schema_identifier] = parsed_schema
            return parsed_schema

        # TODO: Add additional schema types here
        else:
            raise FalseSchemaDefinitionException(schema)

    def struct_to_sql(self, schema: T.StructType, formatted=False) -> str:
        """Convert the given schema into sql rows
        that can form part of a CREATE TABLE statement.
        Includes support for comments, nullability,
        and complex data types (e.g. structs, arrays),

        if formatted is True, the sql will contain newlines
        and be indented for use in a long SQL schema.
        """
        return self._schema_to_spark_sql(schema, formatted=formatted)

    def _schema_to_spark_sql(self, schema: T.StructType, formatted=False) -> str:
        # TODO: Create a more capable method of translating StructTypes to
        # spark sql strings
        # Lacking:
        # - generated-always-as

        rows = []
        for field in schema.fields:
            row = f"{field.name} {field.dataType.simpleString()}"
            if not field.nullable:
                row += " NOT NULL"
            if "comment" in field.metadata:
                # I could have used a repr() here,
                # but then I could get single quoted string. This ensured double quotes
                row += f' COMMENT {json.dumps(field.metadata["comment"])}'
            rows.append(row)

        separator = ",\n  " if formatted else ", "

        str_schema = separator.join(rows)

        return str_schema

    def get_schema_as_string(self, schema_identifier: str) -> str:
        "return schema as a sql schema string"
        schema = self.get_schema(schema_identifier)

        str_schema = self._schema_to_spark_sql(schema)

        return str_schema

    def get_all_schemas(self) -> Dict[str, T.StructType]:
        for id in Configurator().all_keys():
            try:
                self.get_schema(schema_identifier=id)
            except NoSuchSchemaException:
                continue

        return self._registered_schemas

    def get_all_spark_sql_schemas(self) -> Dict[str, str]:
        schemas_dict = self.get_all_schemas()
        str_schemas = {}

        for name, schema in schemas_dict.items():
            str_schema = self._schema_to_spark_sql(schema, formatted=True)
            str_schemas[name] = str_schema

        return str_schemas



import unittest

from spetlr import Configurator
from spetlr.configurator.sql.parse_sql import parse_sql_code_to_config
from spetlr.deltaspec.DeltaDatabaseSpec import DeltaDatabaseSpec


class TestTableSpecConversions(unittest.TestCase):
    def test_DeltaDatabaseSpec_from_sql(self):
        config = parse_sql_code_to_config(
            """
            -- SPETLR.CONFIGURATOR key: MyDbAlias
            CREATE DATABASE IF NOT EXISTS my_db_name
            COMMENT "Really great db"
            LOCATION "/somewhere/over/the/rainbow";
            """
        )

        k, v = config.popitem()
        self.assertEqual(k, "MyDbAlias")
        c = Configurator()
        c.register(k, v)
        db = DeltaDatabaseSpec.from_tc(k)
        self.assertEqual(
            db,
            DeltaDatabaseSpec(
                name="my_db_name",
                location="/somewhere/over/the/rainbow",
                comment="Really great db",
            ),
        )

    def test_DeltaDatabaseSpec_to_sql(self):
        db = DeltaDatabaseSpec(
            name="my_db_name",
            location="/somewhere/over/the/rainbow",
            comment="Really great db",
        )
        self.assertEqual(
            db.get_create_sql(),
            (
                "CREATE SCHEMA IF NOT EXISTS my_db_name\n"
                '  COMMENT "Really great db"\n'
                '  LOCATION "dbfs:/somewhere/over/the/rainbow"'
            ),
        )

    def test_DeltaDatabaseSpec_repr(self):
        db = DeltaDatabaseSpec(
            name="my_db_name",
            location="/somewhere/over/the/rainbow",
            comment="Really great db",
        )
        as_text = repr(db)

        self.assertEqual(db, eval(as_text))

    def test_no_location(self):
        self.assertEqual(
            DeltaDatabaseSpec.from_sql(
                """
            CREATE DATABASE spark_catalog.mydb
            COMMENT "nice data"
        """
            ),
            DeltaDatabaseSpec(name="spark_catalog.mydb", comment="nice data"),
        )

    def test_only_name(self):
        self.assertEqual(
            DeltaDatabaseSpec.from_sql("CREATE DATABASE mydb"),
            DeltaDatabaseSpec(name="mydb"),
        )



from datetime import datetime, timezone
from typing import List

import pyspark.sql.functions as f
from pyspark.sql import DataFrame
from pyspark.sql.types import TimestampType

from spetlr.etl import Transformer
from spetlr.sql.SqlServer import SqlServer
from spetlr.utils import SelectAndCastColumns


class SimpleSqlServerTransformer(Transformer):
    def __init__(
        self,
        *,
        table_id: str,
        server: SqlServer,
        ignoreCase=False,
        dataset_input_keys: List[str] = None,
        dataset_output_key: str = None,
        consume_inputs: bool = True,
    ):
        super().__init__(
            dataset_input_keys=dataset_input_keys,
            dataset_output_key=dataset_output_key,
            consume_inputs=consume_inputs,
        )
        self.server = server
        self.table_id = table_id
        self.ignoreCase = ignoreCase
        self.min_time = datetime(1900, 1, 1, 0, 0, 1, tzinfo=timezone.utc)

    def process(self, df: DataFrame) -> DataFrame:
        # This transformation ensures that the selected columns
        # are casted to the target types
        # [f.col("ColumnName).cast("string").alias("ColumnName"), ...]

        df = SelectAndCastColumns(
            df=df,
            schema=self.server.read_table(self.table_id).schema,
            caseInsensitiveMatching=self.ignoreCase,
        )

        # If the format is timestamp, the seconds should be trunc
        col_choose = [
            (
                f.date_trunc("second", f.col(x[0])).alias(x[0])
                if x[1] == "timestamp"
                else f.col(x[0])
            )
            for x in df.dtypes
        ]
        df = df.select(col_choose)

        # The sql server min date is 1753-01-01. If a timestamp column of the
        # dataframe has a date value lower than this min date, it will not be
        # possible to load the data to the sql server. Here, we check all columns
        # of the dataframe that are the timestamp type, and convert values of
        # those columns to our defined min date, if values are smaller than this date.
        for column in df.columns:
            if df.schema[column].dataType == TimestampType():
                df = df.withColumn(
                    column,
                    f.when(
                        f.col(column) < f.lit(self.min_time), f.lit(self.min_time)
                    ).otherwise(f.col(column)),
                )

        return df



from abc import abstractmethod
from typing import Dict

from pyspark.sql import DataFrame

dataset_group = Dict[str, DataFrame]


class EtlBase:
    def __init__(self):
        super().__init__()

    @abstractmethod
    def etl(self, inputs: dataset_group) -> dataset_group:
        raise NotImplementedError()



import re
from datetime import datetime, timedelta, timezone
from typing import Union

from spetlr import dbg


class PartitionSpec:
    y: Union[None, int]
    m: Union[None, int]
    d: Union[None, int]
    h: Union[None, int]

    def __init__(self, y: int = None, m: int = None, d: int = None, h: int = None):
        self.y = None if y is None else int(y)
        self.m = None if m is None else int(m)
        self.d = None if d is None else int(d)
        self.h = None if h is None else int(h)

    @property
    def ys(self):
        return f"{self.y:04}"

    @property
    def ms(self):
        return f"{self.m:02}"

    @property
    def ds(self):
        return f"{self.d:02}"

    @property
    def hs(self):
        if self.h is None:
            raise ValueError("Hour partition not used")
        return f"{self.h:02}"

    @classmethod
    def from_path(cls, s: str):
        # we got the partition in the form y=2022/m=04/d=04
        # extract the parts
        pattern = r"y=(?P<y>\d+)/m=(?P<m>\d+)/d=(?P<d>\d+)(/h=(?P<h>\d+))?"
        match = re.match(pattern, s)
        return cls(
            y=match.group("y"),
            m=match.group("m"),
            d=match.group("d"),
            h=match.group("h"),  # can be None
        )

    def as_sql_spec(self) -> str:
        s = f"y='{self.ys}', m='{self.ms}', d='{self.ds}'"
        if self.h is not None:
            s += f", h='{self.hs}'"
        return s

    def as_path(self) -> str:
        # returns the partition in the form y=2022/m=04/d=04
        # if some parts are None, they will not appear in the path
        # the partial definition feature assists in discovery
        s = ""
        if self.y is not None:
            s = f"y={self.ys}"
        if self.m is not None:
            s += f"/m={self.ms}"
        if self.d is not None:
            s += f"/d={self.ds}"
        if self.h is not None:
            s += f"/h={self.hs}"
        return s

    def as_datetime(self) -> datetime:
        dbg(
            "returning datetime for "
            f"{repr(self.y)}, {repr(self.m)}, {repr(self.d)}, {repr(self.h)}"
        )
        return datetime(
            year=self.y, month=self.m, day=self.d, hour=self.h or 0, tzinfo=timezone.utc
        )

    def is_earlier_than_dt(self, dt: datetime) -> bool:
        dt = dt.astimezone(tz=timezone.utc)
        dt = dt.replace(minute=0, second=0, microsecond=0)
        if self.h is None:
            dt.replace(hour=0)
        return self.as_datetime() < dt

    def next(self):
        dt = self.as_datetime()
        if self.h is None:
            delta = timedelta(days=1)
        else:
            delta = timedelta(hours=1)

        next_dt = dt + delta

        return self.__class__(
            y=next_dt.year,
            m=next_dt.month,
            d=next_dt.day,
            h=next_dt.hour if self.h is not None else None,
        )



import unittest
from typing import List, Tuple

from spetlrtools.testing import DataframeTestCase

from spetlr import Configurator
from spetlr.delta import DbHandle, DeltaHandle
from spetlr.etl import Orchestrator
from spetlr.etl.extractors.stream_extractor import StreamExtractor
from spetlr.etl.loaders.upsert_loader_streaming import UpsertLoaderStreaming
from spetlr.spark import Spark
from spetlr.testutils.stop_test_streams import stop_test_streams
from spetlr.utils import DataframeCreator
from tests.cluster.delta import extras
from tests.cluster.delta.SparkExecutor import SparkSqlExecutor


@unittest.skipUnless(
    Spark.version() >= Spark.DATABRICKS_RUNTIME_10_4,
    f"UpsertLoader for streaming not available for Spark version {Spark.version()}",
)
class UpsertLoaderTestsDeltaStream(DataframeTestCase):
    target_dh: DeltaHandle
    target_id = "UpsertLoaderStreamingTarget"
    source_id = "UpsertLoaderStreamingSource"

    join_cols = ["col1", "col2"]

    data1 = [
        (5, 6, "foo"),
        (7, 8, "bar"),
    ]
    data2 = [
        (1, 2, "baz"),
    ]
    data3 = [(5, 6, "boo"), (5, 7, "spam")]
    # data5 is the merge result of data2 + data3 + data4
    data4 = [(1, 2, "baz"), (5, 6, "boo"), (5, 7, "spam"), (7, 8, "bar")]

    dummy_columns: List[str] = ["col1", "col2", "col3"]

    @classmethod
    def setUpClass(cls) -> None:
        tc = Configurator()
        tc.add_resource_path(extras)
        tc.set_debug()

        tc.register(
            "MyStream",
            {
                "query_name": "testquerytbl{ID}",
            },
        )

        SparkSqlExecutor().execute_sql_file("upsertloader-test")

        cls.source_dh = DeltaHandle.from_tc(cls.source_id)
        cls.target_dh = DeltaHandle.from_tc(cls.target_id)

        cls.dummy_schema = cls.target_dh.read().schema

    @classmethod
    def tearDownClass(cls) -> None:
        DbHandle.from_tc("UpsertLoaderDb").drop_cascade()

        stop_test_streams()

    def test_01_can_perform_incremental_on_empty(self):
        """Stream two rows to the empty target table"""

        self._create_test_source_data(data=self.data1)

        self.execute_upsert_stream_orchestrator()

        self.assertDataframeMatches(self.target_dh.read(), None, self.data1)

    def test_02_can_perform_incremental_append(self):
        """The target table is already filled from before.
        One new rows appear in the source table to be streamed
        """

        existing_rows = self.target_dh.read().collect()
        self.assertEqual(2, len(existing_rows))

        self._create_test_source_data(data=self.data2)

        self.execute_upsert_stream_orchestrator()

        self.assertDataframeMatches(
            self.target_dh.read(), None, self.data1 + self.data2
        )

    def test_03_can_perform_merge(self):
        """The target table is already filled from before.
        Two new rows appear in the source table - one of them will be merged.
        """
        existing_rows = self.target_dh.read().collect()
        self.assertEqual(3, len(existing_rows))

        self._create_test_source_data(data=self.data3)

        self.execute_upsert_stream_orchestrator()

        self.assertDataframeMatches(self.target_dh.read(), None, self.data4)

    def _create_test_source_data(
        self, tableid: str = None, data: List[Tuple[int, int, str]] = None
    ):
        if tableid is None:
            tableid = self.source_id
        if data is None:
            raise ValueError("Testdata missing.")

        dh = DeltaHandle.from_tc(tableid)

        Spark.get().sql(
            f"""
            CREATE TABLE IF NOT EXISTS {dh.get_tablename()}
            (
            id int,
            name string
            )
            LOCATION '{Configurator().get(tableid,"path")}'
            """
        )

        df_source = DataframeCreator.make_partial(
            self.dummy_schema, self.dummy_columns, data
        )

        dh.append(df_source)

    def execute_upsert_stream_orchestrator(self):
        o = Orchestrator()
        o.extract_from(StreamExtractor(self.source_dh, dataset_key="MyTbl"))
        o.load_into(
            UpsertLoaderStreaming(
                handle=self.target_dh,
                options_dict={},
                trigger_type="availablenow",
                checkpoint_path=Configurator().get(self.target_id, "checkpoint_path"),
                await_termination=True,
                upsert_join_cols=self.join_cols,
                query_name=Configurator().get("MyStream", "query_name"),
            )
        )
        o.execute()



from typing import List

from pyspark.sql import DataFrame

from spetlr.etl import Transformer
from spetlr.utils import DropOldestDuplicates


class DropOldestDuplicatesTransformer(Transformer):
    def __init__(
        self,
        *,
        cols: List[str],
        orderByColumn: str,
        dataset_input_keys: List[str] = None,
        dataset_output_key: str = None,
        consume_inputs: bool = True,
    ):
        super().__init__(
            dataset_input_keys=dataset_input_keys,
            dataset_output_key=dataset_output_key,
            consume_inputs=consume_inputs,
        )
        self.cols = cols
        self.orderByColumn = orderByColumn

    def process(self, df: DataFrame) -> DataFrame:
        return DropOldestDuplicates(
            df=df, cols=self.cols, orderByColumn=self.orderByColumn
        )



from abc import abstractmethod

from pyspark.sql import DataFrame

from spetlr.etl import EtlBase, dataset_group


class LazyExtractor(EtlBase):
    """In some cases extractors depend on other extractors,
    for these situations, the datasets from previous steps
    are accessible in the variable self.previous_extractions.

    In regards to the etl step, the extractor ADDs its dataset
    to the total set of datasets.

    This Lazy Extractor is very similar to the standard extractor
    with the key difference that it will only read its target
    if the dataset key is previously missing.
    """

    def __init__(self, dataset_key: str = None):
        super().__init__()
        self.dataset_key = dataset_key or type(self).__name__
        self.previous_extractions: dataset_group = {}

    def etl(self, inputs: dataset_group) -> dataset_group:
        self.previous_extractions = inputs
        if self.dataset_key not in inputs:
            inputs[self.dataset_key] = self.read()

        return inputs

    @abstractmethod
    def read(self) -> DataFrame:
        raise NotImplementedError()



from datetime import datetime, timezone
from typing import List

import pyspark.sql.functions as f
from pyspark.sql import DataFrame, Window

from spetlr.etl import Transformer
from spetlr.utils import DropOldestDuplicates


class ValidFromToTransformer(Transformer):
    """
    The class introduces SCD2 columns:
        ValidFrom
        ValidTo
        IsCurrent
    NB: Be aware, if incremental extraction is used, the logic does not work
    """

    def __init__(
        self,
        time_col: str,
        wnd_cols: List[str],
        dataset_input_keys: List[str] = None,
        dataset_output_key: str = None,
        consume_inputs: bool = True,
    ):
        super().__init__(
            dataset_input_keys=dataset_input_keys,
            dataset_output_key=dataset_output_key,
            consume_inputs=consume_inputs,
        )
        self._time_col = time_col
        self._wnd_cols = wnd_cols

    def process(self, df: DataFrame) -> DataFrame:
        _cols = df.columns
        df = df.withColumn("ValidFrom", f.col(self._time_col))

        # Drop duplicates based on the window cols and time col
        # If there is duplicates,
        # the ValidFromTransformer could introduce multiple is_current
        df = DropOldestDuplicates(
            df=df,
            cols=self._wnd_cols + [self._time_col],
            orderByColumn=self._time_col,
        )

        # The largest year number allowed in python is 9999
        # but using this value can cause some pyspark errors
        # When .collect() the data, the pyspark/sql/types.py
        # find it as an "Invalid argument"

        # Therefore, this class uses the max time from pandas
        # as it seems to align with pyspark
        # https://pandas.pydata.org/docs/reference/api/pandas.Timestamp.max.html

        max_time = datetime(2262, 4, 11, tzinfo=timezone.utc)

        return (
            df.withColumn(
                "NextValidFrom",
                f.lead("ValidFrom").over(
                    Window.partitionBy(self._wnd_cols).orderBy("ValidFrom")
                ),
            )
            .withColumn("ValidTo", f.coalesce(f.col("NextValidFrom"), f.lit(max_time)))
            .withColumn("IsCurrent", f.col("NextValidFrom").isNull())
            .drop("NextValidFrom")
            .select(*_cols, "ValidFrom", "ValidTo", "IsCurrent")
        )



import unittest
import uuid

import pyspark.sql.types as T
from spetlrtools.testing import DataframeTestCase

from spetlr import Configurator
from spetlr.delta.delta_handle import DeltaHandleInvalidFormat
from spetlr.filehandle import FileHandle
from spetlr.schema_manager import SchemaManager


class FileHandleTests(DataframeTestCase):
    path_unique_id = uuid.uuid4().hex

    data_format = "json"
    file_location = f"file_location/{path_unique_id}/"
    checkpoint_path = f"checkpoint_location/{path_unique_id}/"
    schema_location = f"schema_location/{path_unique_id}/"

    options = {"test_option": "test_value"}

    cloud_files_options = {
        "cloudFiles.format": data_format,
        "cloudFiles.schemaLocation": schema_location,
    }

    schema = T.StructType(
        [
            T.StructField("id", T.IntegerType(), True),
            T.StructField("name", T.StringType(), True),
        ]
    )

    @classmethod
    def setUpClass(cls) -> None:
        sc = SchemaManager()
        sc.clear_all_configurations()

        sc.register_schema("SCHEMA", cls.schema)

        tc = Configurator()
        tc.clear_all_configurations()

        tc.register(
            "FileSource",
            {
                "name": "TestFileSource",
                "path": cls.file_location,
                "format": cls.data_format,
                "schema_location": cls.schema_location,
                "schema": "SCHEMA",
            },
        )

    def test_01_create_file_handle(self):
        handle = FileHandle(
            file_location=self.file_location,
            schema_location=self.schema_location,
            data_format=self.data_format,
            options=self.options,
            schema=self.schema,
        )

        self.assertEquals(handle._location, self.file_location)
        self.assertEquals(handle._data_format, self.data_format)
        self.assertEquals(handle._schema_location, self.schema_location)
        self.assertEquals(handle._options, {**self.options, **self.cloud_files_options})
        self.assertEqualSchema(handle._schema, self.schema)

    def test_02_create_file_handle_from_tc(self):
        handle = FileHandle.from_tc("FileSource")

        self.assertEquals(handle._location, self.file_location)
        self.assertEquals(handle._data_format, self.data_format)
        self.assertEquals(handle._schema_location, self.schema_location)
        self.assertEquals(handle._options, self.cloud_files_options)
        self.assertEqualSchema(handle._schema, self.schema)

    def test_03_validate_wrong_format_exception(self):
        with self.assertRaises(DeltaHandleInvalidFormat):
            FileHandle(
                file_location=self.file_location,
                schema_location=self.schema_location,
                data_format="delta",
                options=self.options,
                schema=self.schema,
            )


if __name__ == "__main__":
    unittest.main()
